#include "TamaP.h"
#include "math.h"

char * TamaClass_GetName(TamaClass tama_class)
{ return (tama_class->name); }

double TamaClass_GetProbability(TamaClass tama_class)
{ return (tama_class->probability); }

ObjList TamaClass_GetGCListList(TamaClass tama_class)
{ return (tama_class->gc_list_list); }

ObjList TamaClass_GetNextList(TamaClass tama_class)
{ return (tama_class->next_list); }

double TamaNext_GetPower(TamaNext tama_next)
{ return (tama_next->power); }

TamaClass Tamas_GetTamaClass(Tamas tamas)
{ return (tamas->tama_class); }

double * Tamas_GetX(Tamas tamas) { return (tamas->x); }
double * Tamas_GetY(Tamas tamas) { return (tamas->y); }
double * Tamas_GetVx(Tamas tamas) { return (tamas->vx); }
double * Tamas_GetVy(Tamas tamas) { return (tamas->vy); }
int Tamas_GetNumber(Tamas tamas) { return (tamas->number); }

int TamaNext_GetNumber(TamaNext tama_next)
{ return (tama_next->number); }

TamaClass TamaNext_GetTamaClass(TamaNext tama_next)
{ return (tama_next->tama_class); }

/*===========================================================================*/
/* ƥ֥Ȥκ                                              */
/*===========================================================================*/

TamaClass TamaClass_Create(char * name,
			   int size,
			   double probability,
			   double air,
			   int after_image_length,
			   ObjList gc_list_list,
			   ObjList next_list, Disp disp)
{
  TamaClass tama_class;

  tama_class = (TamaClass)malloc(sizeof(_TamaClass));
  if (!tama_class) Error("TamaClass_Create", "Cannot allocate memory.");

  if ((name == NULL) || (name[0] == '\0')) name = "Unknown\0";
  tama_class->name = strdup(name);
  if (tama_class->name == NULL)
    Error("TamaClass_Create", "Cannot allocate memory.");
  tama_class->size = size;
  tama_class->probability = probability;
  tama_class->air = air;
  tama_class->after_image_length = after_image_length;

  if (gc_list_list) tama_class->gc_list_list = gc_list_list;
  else              tama_class->gc_list_list = ObjList_Create();

  if (next_list) tama_class->next_list = next_list;
  else           tama_class->next_list = ObjList_Create();

  return (tama_class);
}

TamaNext TamaNext_Create(TamaClass tama_class, double power, int n)
{
  TamaNext next;

  next = (TamaNext)malloc(sizeof(_TamaNext));
  if (!next) Error("TamaNext_Create", "Cannot allocate memory.");

  next->tama_class = tama_class;
  next->power = power;
  next->number = n;

  return (next);
}

TamaClass TamaClass_Destroy(TamaClass tama_class)
{
  if (!tama_class) return (NULL);

  if (tama_class->name) free(tama_class->name);

  if (tama_class->gc_list_list)
    ObjList_Destroy(tama_class->gc_list_list);

  if (tama_class->next_list)
    ObjList_Destroy(tama_class->next_list);

  free(tama_class);

  return (NULL);
}

TamaNext TamaNext_Destroy(TamaNext next)
{
  if (!next) return (NULL);

  free(next);

  return (NULL);
}

/*===========================================================================*/
/* ǻѤؿ                                                        */
/*===========================================================================*/

/* TamaClass θ */

static TamaClass TamaClassList_SearchOrCreateTamaClass(ObjList list,
						       char * name,
						       Disp disp,
						       int default_size,
						       double default_air,
						       int default_after_image)
{
  ObjListData current;
  TamaClass tama_class;

  if (!list) Error("TamaClassList_SearchOrCreateTamaClass",
		   "TamaClassList has not created.");

  for (current = ObjList_GetStart(list);
       !ObjList_IsEndEdge(list, current);
       current = ObjListData_GetNext(current)) {
    tama_class = (TamaClass)ObjListData_GetObj(current);
    if (!strcmp(tama_class->name, name)) {
      return (tama_class);
    }
  }

  /* ĤʤäˤϡꥹȤɲä */
  tama_class = TamaClass_Create(name,
				default_size,
				-1.0,
				default_air,
				default_after_image,
				NULL, NULL, disp);
  ObjList_InsertObjToEnd(list, tama_class, (ObjDestructor)TamaClass_Destroy);

  return (tama_class);
}

/*---------------------------------------------------------------------------*/
/* ñμ                                                                */
/*---------------------------------------------------------------------------*/

static char * GetWord(Stream stream)
{
  char * word;
  word = Stream_GetWord(stream, NULL, 0,
			"",     /* ʸڤʬ */
			" \t",  /* ʸǤڤʬ */
			"!#$%", /* ȹԤƬʸ '#' ʤ */
			"\n",   /* ʸ '\n' ʤ */
			"\"\'", /* ʸ '\"' ʤ */
			"",     /* ʸƬɤߤȤФʸ */
			""      /* ʸɤߤȤФʸ */
			);
  return (word);
}

static char * GetWord2(Stream stream)
{
  char * word;
  word = Stream_GetWord(stream, NULL, 0,
			"",     /* ʸڤʬ */
			" \t",  /* ʸǤڤʬ */
			"",     /* ȹԤƬʸ '#' ʤ */
			"\n",   /* ʸ '\n' ʤ */
			"\"\'", /* ʸ '\"' ʤ */
			"",     /* ʸƬɤߤȤФʸ */
			""      /* ʸɤߤȤФʸ */
			);
  return (word);
}

static ObjList ReadNext(ObjList tama_class_list, char * word, Disp disp,
			int default_size,
			double default_air,
			int default_after_image)
{
  ObjList list;
  Stream stream;
  char * w1;
  char * w2;
  char * w3;
  TamaClass t;
  double power;
  int n;

  list = ObjList_Create();
  if (!word) return (list);
  stream = Stream_CreateFromCharacters(word);
  if (!stream) return (list);

  while ((w1 = GetWord2(stream)) != NULL) {
    t = TamaClassList_SearchOrCreateTamaClass(tama_class_list, w1, disp,
					      default_size, default_air,
					      default_after_image);
    w2 = GetWord2(stream);
    if ((w3 = GetWord2(stream)) != NULL) {
      power = atof(w2);
      n = atoi(w3);
      ObjList_InsertObjToEnd(list, TamaNext_Create(t, power, n),
			     (ObjDestructor)TamaNext_Destroy);
    }
    if (w1) free(w1);
    if (w2) free(w2);
    if (w3) free(w3);
  }

  Stream_Destroy(stream);

  return (list);
}

/* ǡɤ߹ */
ObjList TamaClassList_CreateFromStream(Disp disp,
				       ColorGCList color_gc_list,
				       Stream stream, char * background_color)
{
  ObjList tama_class_list;
  TamaClass current = NULL;
  char * word;
  char * word2;
  Stream s;
  double default_probability;
  int default_size;
  double default_air;
  int default_after_image;

  tama_class_list = ObjList_Create();

  while ((word = GetWord(stream)) != NULL) {
    word2 = NULL;

    if (!strcmp(word, "Name")) {
      if ((word2 = GetWord(stream)) != NULL) {
	current =
	  TamaClassList_SearchOrCreateTamaClass(tama_class_list, word2, disp,
						default_size, default_air,
						default_after_image);
      }
    } else if (!strcmp(word, "DefaultProbability")) {
      if ((word2 = GetWord(stream)) != NULL) {
	default_probability = atof(word2);
      }
    } else if (!strcmp(word, "DefaultSize")) {
      if ((word2 = GetWord(stream)) != NULL) {
	default_size = atoi(word2);
      }
    } else if (!strcmp(word, "DefaultAir")) {
      if ((word2 = GetWord(stream)) != NULL) {
	default_air = atof(word2);
      }
    } else if (!strcmp(word, "DefaultAfterImage")) {
      if ((word2 = GetWord(stream)) != NULL) {
	default_after_image = atoi(word2);
      }
    } else {
      if (current) {
	if (!strcmp(word, "Size")) {
	  if ((word2 = GetWord(stream)) != NULL) {
	    if (!strcmp(word2, "Default")) current->size = default_size;
	    else current->size = atoi(word2);
	  }
	} else if (!strcmp(word, "Probability")) {
	  if ((word2 = GetWord(stream)) != NULL) {
	    if      (!strcmp(word2, "Default"))
	      current->probability = default_probability;
	    else if (!strcmp(word2, "None")) current->probability = -1.0;
	    else current->probability = atof(word2);
	  }
	} else if (!strcmp(word, "Air")) {
	  if ((word2 = GetWord(stream)) != NULL) {
	    if (!strcmp(word2, "Default")) current->air = default_air;
	    else current->air = atof(word2);
	  }
	} else if (!strcmp(word, "AfterImage")) {
	  if ((word2 = GetWord(stream)) != NULL) {
	    if (!strcmp(word2, "Default"))
	      current->after_image_length = default_after_image;
	    else current->after_image_length = atoi(word2);
	  }
	} else if (!strcmp(word, "Color")) {
	  if ((word2 = GetWord(stream)) != NULL) {
	    if (current->gc_list_list == NULL)
	      current->gc_list_list = ObjList_Create();
	    ObjList_Concatenate(current->gc_list_list,
				CreateColorGCListFromCharacters(disp,
								color_gc_list,
								word2,
								background_color,
								current->after_image_length));

	  }
	} else if (!strcmp(word, "Next")) {
	  if ((word2 = GetWord(stream)) != NULL) {
	    if (current->next_list == NULL)
	      current->next_list = ObjList_Create();
	    if (strcmp(word2, "None"))
	      ObjList_Concatenate(current->next_list,
				  ReadNext(tama_class_list, word2, disp,
					   default_size,
					   default_air,
					   default_after_image));
	  }
	} else if (!strcmp(word, "End")) {
	  if (word)  free(word);
	  if (word2) free(word2);
	  return (tama_class_list);
	} else {
	  fprintf(stderr, "Unknown command : %s\n", word);
	}
      } else {
	fprintf(stderr,
		"Cannot specify class name. Command was ignored. : %s\n",
		word);
      }
    }
    if (word)  free(word);
    if (word2) free(word2);
  }

  return (tama_class_list);
}

/*===========================================================================*/
/* Tamas ֥Ȥ                                                  */
/*===========================================================================*/

/*---------------------------------------------------------------------------*/
/* Tamas ֥Ȥ                                                  */
/*---------------------------------------------------------------------------*/

Tamas Tamas_Create(TamaClass tama_class, double power, int number, int n,
		   double * x, double * y,
		   double * vx, double * vy,
		   int x_min, int y_min, int x_max, int y_max)
{
  double rad;
  Tamas tamas;
  double p;
  int i, j, a;

  tamas = (Tamas)malloc(sizeof(_Tamas));
  if (!tamas) Error("Tamas_Create", "Cannot allocate memory");

  tamas->tama_class = tama_class;

  tamas->x  = (double *)malloc(sizeof(double) * number * n);
  if (tamas->x  == NULL) Error("Tamas_Create", "Cannot allocate memory");

  tamas->y  = (double *)malloc(sizeof(double) * number * n);
  if (tamas->y  == NULL) Error("Tamas_Create", "Cannot allocate memory");

  tamas->vx = (double *)malloc(sizeof(double) * number * n);
  if (tamas->vx == NULL) Error("Tamas_Create", "Cannot allocate memory");

  tamas->vy = (double *)malloc(sizeof(double) * number * n);
  if (tamas->vy == NULL) Error("Tamas_Create", "Cannot allocate memory");

  tamas->gc_list = ObjList_GetStart(tamas->tama_class->gc_list_list);

  a = 0;
  for (i = 0; i < n; i++) {
    for (j = 0; j < number; j++) {
      p = cos(DoubleRand(M_PI * 0.5));
      rad = DoubleRand(M_PI * 2.0);
      tamas->x[a] = x[i];
      tamas->y[a] = y[i];
      tamas->vx[a] = cos(rad) * power * p + vx[i];
      tamas->vy[a] = sin(rad) * power * p + vy[i];
      if ( ((tamas->x[a] < x_min) && (tamas->vx[a] < 0.0)) ||
	   ((tamas->x[a] > x_max) && (tamas->vx[a] > 0.0)) ||
	   ((tamas->y[a] > y_max) && (tamas->vy[a] > 0.0)) ) {
	/* None */
      } else {
	a++;
      }
    }
  }
  tamas->number = a;

  return (tamas);
}

/*---------------------------------------------------------------------------*/
/* Tama ֥Ȥκ                                                   */
/*---------------------------------------------------------------------------*/

Tamas Tamas_Destroy(Tamas tamas)
{
  if (!tamas) return (NULL);

  if (tamas->x ) free(tamas->x);
  if (tamas->y ) free(tamas->y);
  if (tamas->vx) free(tamas->vx);
  if (tamas->vy) free(tamas->vy);

  free(tamas);

  return (NULL);
}

/*---------------------------------------------------------------------------*/
/* Tamas ֥Ȥΰư                                                  */
/*---------------------------------------------------------------------------*/

AfterImages Tamas_Move(Tamas tamas, int x_min, int y_min, int x_max, int y_max,
		       int size)
{
  AfterImages after_images;
  int length;
  int i;

  if (ObjList_IsEndEdge(tamas->tama_class->gc_list_list, tamas->gc_list))
    return (NULL);

  for (i = 0; i < tamas->number; i++) {

    tamas->vy[i] += GRAVITY;

    /* 񹳤η׻ */
    tamas->vx[i] *= (1.0 - tamas->tama_class->air);
    tamas->vy[i] *= (1.0 - tamas->tama_class->air);

    tamas->x[i] += (double)tamas->vx[i] * size / 1000;
    tamas->y[i] += (double)tamas->vy[i] * size / 1000;

    if (((tamas->x[i] < x_min) && (tamas->vx[i] < 0.0)) ||
	((tamas->x[i] > x_max) && (tamas->vx[i] > 0.0)) ||
	(tamas->y[i] > y_max)) {
    }
  }

  after_images = AfterImages_Create(tamas->tama_class->size, tamas->number,
				    tamas->x, tamas->y,
				    x_min, y_min, x_max, y_max,
				    ObjListData_GetObj(tamas->gc_list));

  tamas->gc_list = ObjListData_GetNext(tamas->gc_list);

  return (after_images);
}
